//copy from fengmap2.2
/**
 * FMHeatMap 热力图, 用于地图热力展示使用。
 * @class
 * @hideconstructor true
 * @memberof fengmap
 */

function FMHeatMap(opts) {
    
    this.texture_ = null;

    this.config(opts);
}

/**
* 使用create方法创建热力图对象

* @param  {fengmap.FMMap} map map对象
* @param  {object} config 配置项
*             @param {object} config.radius  热点半径
*             @param {string} config.opacity 热力图透明度，范围0-1
*             @param {string} config.max  热力点value的最大值
*             @param {int} config.maxSize  热力图长宽的最大尺寸
*             @param {boolean} config.gradient  热力图的渐变颜色。默认为：{ 0.45: "rgb(0,0,255)", 0.55: "rgb(0,255,255)", 0.65: "rgb(0,255,0)", 0.95: "yellow", 1.0: "rgb(255,0,0)" }
* @static
* @return {fengmap.FMHeatMap}
*/
FMHeatMap.create = function (config) {

    return new FMHeatMap(config);

};

FMHeatMap.prototype = {

    //获取屏幕DPI
    getDeviceDPI: function () {
        let arrDPI = [];
        if (window.screen.deviceXDPI !== undefined) {
            arrDPI[0] = window.screen.deviceXDPI;
            arrDPI[1] = window.screen.deviceYDPI;
        } else {
            let tmpNode = document.createElement("DIV");
            tmpNode.style.cssText = "width:1in;height:1in;position:absolute;left:0px;top:0px;z-index:99;visibility:hidden";
            document.body.appendChild(tmpNode);
            arrDPI[0] = parseInt(tmpNode.offsetWidth);
            arrDPI[1] = parseInt(tmpNode.offsetHeight);
            tmpNode.parentNode.removeChild(tmpNode);
        }
        return arrDPI;
    },
    //获取地图范围的高宽。单位像素
    getDeviceWH: function (opts) {
        opts = opts || {};
        let w = opts.width;
        let h = opts.height;
        let wh = {
            w: Math.round(w),
            h: Math.round(h)
        };

        let dpi = this.getDeviceDPI();

        // wh.w = fm.Math.nextPowerOfTwo(wh.w * dpi[0] / 25.4); //取距离最近的2的幂，否则会输出警告
        // wh.h = fm.Math.nextPowerOfTwo(wh.h * dpi[1] / 25.4);

        // 使用 nearest 不使用上面的 next, 不浪费空间
        // wh.w = fm.Math.nearestPowerOfTwo(wh.w * dpi[0] / 25.4); //取距离最近的2的幂，否则会输出警告
        // wh.h = fm.Math.nearestPowerOfTwo(wh.h * dpi[1] / 25.4);

        // 使用 上面的方法得到的 2的幂的大小, 会使得最后的结果点不"圆"
//        wh.w = parseInt(wh.w * dpi[0] / 25.4);
//        wh.h = parseInt(wh.h * dpi[1] / 25.4);

        /////////////////////////////////
        // 20161201 clamp size to 2048 //
        /////////////////////////////////
        let rat = 1;
        let maxSize = opts.maxSize || 2048;
        if (wh.w >= maxSize || wh.h >= maxSize) {
            if (wh.w >= wh.h) {
                rat = wh.h / wh.w;
                wh.w = maxSize;
                wh.h = parseInt(maxSize * rat);
            } else {
                rat = wh.w / wh.h;
                wh.h = maxSize;
                wh.w = parseInt(maxSize * rat);
            }
        }

        return wh;
    },

    //转换地图坐标到像素坐标
    getPixXY: function (x, y) {

        let aa = (x - this.map_.minX) / (this.map_.maxX - this.map_.minX);
        let bb = (this.map_.maxY - y) / (this.map_.maxY - this.map_.minY);

        let pnt = {
            x: aa,
            y: bb
        };

        pnt.x = pnt.x * this.options.width;
        pnt.y = pnt.y * this.options.height;

        return pnt;
    },

    // 初始化参数配置
    config: function (opts) {
        let me = this;

        let wh = me.getDeviceWH(opts);

        let w = wh.w,
            h = wh.h;

        me.options = {
            width: w,
            height: h,
            // 热图画布透明度，取值范围 0-1
            opacity: opts.opacity ? opts.opacity * 255 : 100 * 255,
            // 点半径
            radius: opts.radius ? opts.radius : 30,
            // 边界模糊半径
            bshadow: opts.radius ? opts.radius / 20 : 30 / 20,
            maxSize: opts.mapSize || 2048,
            boundVal: 15000,
            shadowBlur: Math.ceil(opts.radius * 1),//0.3
            // 缓存画布中点数据，并且记录最大值, 当最大值发生变化的时候，热图画布需要重新根据缓存的点得数据绘制
            points: {
                max: opts.max ? opts.max : 100,
                data: []
            },
            // 调色板颜色取值范围
            gradient: opts.gradient ? opts.gradient : {
                0.45: "rgb(0,0,255)",
                0.55: "rgb(0,255,255)",
                0.65: "rgb(0,255,0)",
                0.95: "yellow",
                1.0: "rgb(255,0,0)"
            },
            isHeatPlan: opts.isHeatPlan,
            heatHeight: opts.heatHeight,
        };

        // 热图画布
        //let container = document.getElementById(id);
        let canvas = document.createElement("canvas");
        let ctx = canvas.getContext('2d');

        canvas.width = w;
        canvas.height = h;

        me.options.ctx = ctx;
        me.options.canvas = canvas;

        // console.log(w,h);
    },

    // 绘制圆阴影：可视区域只绘制了一个圆阴影效果，而圆本身的位置在可视区域之外
    renderShadow: function (x, y, val, cache) {

        let me = this,
            ctx = me.options.ctx,
            radius = me.options.radius,
            bval = me.options.boundVal;

        let opacity = parseFloat(val / me.options.points.max, 10);

        ctx.save();

        // ctx.globalAlpha = 0.1;
        ctx.shadowColor = ('rgba(0, 0, 0, ' + opacity + ')');
        ctx.shadowOffsetX = bval;
        ctx.shadowOffsetY = bval;
        ctx.shadowBlur = me.options.shadowBlur;
        ctx.beginPath();
        ctx.arc(x - bval, y - bval, me.options.radius, 0, Math.PI * 2, true);
        ctx.closePath();
        ctx.fill();

        ctx.restore();

        if (!cache) {
            me.cachePoint(x, y, val);
        }

    },

    // 绘制彩色热图根据调色板
    colorize: function () {

        let me = this,
            w = me.options.width,
            h = me.options.height,
            ctx = me.options.ctx;
        let img = ctx.getImageData(0, 0, w, h);

        let imgData = img.data,
            len = imgData.length,
            palette = me.getPalette();
        let opacity = me.options.opacity;

        for (let i = 3; i < len; i += 4) {
            // [0] -> r, [1] -> g, [2] -> b, [3] -> alpha
            let alpha = imgData[i],
                // 是因为要取调色板中的颜色，而调色板数的颜色数据对应的步长是4
                offset = alpha * 4;

            if (!offset) {
                continue;
            }
            offset = Math.floor(offset);

            let finalAlpha = opacity < alpha ? opacity : alpha;
            if (finalAlpha < 0.8) {
                finalAlpha = 0.8;
            }

            imgData[i - 3] = palette[offset];
            imgData[i - 2] = palette[offset + 1];
            imgData[i - 1] = palette[offset + 2];
            imgData[i] = finalAlpha;

        }

        ctx.putImageData(img, 0, 0);

    },

    //地图上显示热力图
    renderToMap: function (groupID) {
        let me = this;
        let canvas = me.options.canvas;

        let texture_ = new fm.Texture(canvas);
        this.setTexture(groupID, texture_);

    },

    //从地图上移除热力图
    disposeHeatMap: function (groupID) {
        if (!groupID) return;
        let me = this;

        let floor = this.map_.getFMGroup(groupID).getExtentMeshs()[0];

        if (!floor._hasHeatMap) {
            return;
        }

        // 加一个标签,记录当前的heapMap状态
        floor._hasHeatMap = false;

        //
        // release memory from gl
        //
        if (floor.material.map) {
            floor.material.map.dispose();
        }
        if (this.options.isHeatPlan) {

            floor.material.map = null;
            floor.material.mapMixColor = false;
            floor.material.transparent = true;
            floor.material.needsUpdate = true;
            this.map_.getFMGroup(groupID)._layers.model[0]._scene.getObjectByName("directionalLight").remove(this.map_.getFMGroup(groupID)._layers.model[0]._scene.getObjectByName("heatMap"));

        } else {
            let ms = this.map_.getFMGroup(groupID).getModelMeshs();
            this.map_.getFMGroup(groupID)._layers.model[0]._scene.traverse((mesh) => {
                if (mesh.name === "heatmap") {
                    mesh.parent.remove(mesh);
                }
            });
            ms.forEach(function (itm) {

                if (itm.material.type === "MultiMaterial") {
                    // support multiMaterial index
                    let matsLen = itm.material.materials.length;
                    let matIndex = 0,
                        step = 3;

                    while (matIndex < matsLen) {
                        let _mat = itm.material.materials[matIndex];

                        _mat.map = null;

                        // _mat.color.copy(_mat._color);
                        _mat.mapMixColor = false;
                        _mat.transparent = true;
                        _mat.needsUpdate = true;

                        step = matIndex === 0 ? 3 : 2;
                        matIndex += step;
                    }

                } else {
                    //lazyCreateModel
                    let _mat = itm.material;
                    _mat.map = null;

                    // _mat.color.copy(_mat._color);
                    _mat.mapMixColor = false;
                    _mat.transparent = true;
                    _mat.needsUpdate = true;
                }
            });
        }


    },


    // 通过调色板来获取平滑的颜色值
    getPalette: function () {
        // 调色板画布
        let pcanvas = document.createElement("canvas"),
            pctx = pcanvas.getContext('2d');
        pcanvas.width = 1;
        pcanvas.height = 256;

        let me = this,
            gradient = me.options.gradient;

        //缓存调色板数据
        let grad = pctx.createLinearGradient(0, 0, 1, 256);
        for (let x in gradient) {
            grad.addColorStop(x, gradient[x]);
        }
        pctx.fillStyle = grad;
        pctx.fillRect(0, 0, 1, 256);
        // 这个有说明: http://www.w3school.com.cn/tags/canvas_getimagedata.asp
        // 返回的是一个一位数组，每一个像素的的颜色有四个值来表示
        // 前三个值表示 红绿蓝，第四个值表示alpha 通道
        // 也就是这个一位数组的长度是: 1 * 256 * 4;
        return pctx.getImageData(0, 0, 1, 256).data;
    },

};

// 缓存点数据
FMHeatMap.prototype.cachePoint = function (x, y, val) {
    let me = this,
        points = me.options.points,
        data = points.data;
    if (val > points.max) {
        // points.max = val;
    }

    data.push([x, y, val]);
};

/**
 * 清除热力图的所有热力点
 */
FMHeatMap.prototype.clearPoints = function () {
    this.options.points.data = [];
    this.clear();
};

/**
 * 增加一个热点
 *         @param {float} x  地图坐标x值。
 *         @param {float} y  地图坐标y值。
 *         @param {float} value 热点所占的比重值
 */
// 增加一个热点
FMHeatMap.prototype.addPoint = function (x, y, value) {

    if (!x || !y || !value) return;
    // let newXY = this.getPixXY(x, y);
    newXY.x=x;
    newXY.y=y;
    if (!newXY) return;

    let me = this;
    me.clear();

    let data = me.options.points.data,
        len = data.length;

    for (let i = 0; i < len; i++) {
        me.renderShadow(data[i][0], data[i][1], data[i][2], true);
    }

    me.renderShadow(newXY.x, newXY.y, value, false);
    me.colorize();

};

//
// 根据给定的点
//
FMHeatMap.prototype.getTexture = function (points) {

    this.clearPoints();
    this.addPoints(points);

    let clone = this.options.canvas.cloneNode();
    clone.getContext('2d').drawImage(this.options.canvas, 0, 0);
  
    return clone.toDataURL();

};

FMHeatMap.prototype.clear = function () {
    let ctx = this.options.ctx;
    ctx.clearRect(0, 0, this.options.width, this.options.height);
};

/**
 * 批量增加热点
 * @param  {array} datas datas数组的每一个值是{x,y,value}对象。
 */
FMHeatMap.prototype.addPoints = function (points) {
    let me = this;
    // me.options.ctx.clearRect(0, 0, me.options.width, me.options.height);
    me.clear();

    let data = me.options.points.data;
    let len = data.length;

    for (let i = 0; i < len; i++) {
        me.renderShadow(data[i][0], data[i][1], data[i][2], true);
    }

    let plen = points.length;
    for (let j = 0; j < plen; j++) {
        // let newxy = this.getPixXY(points[j].x, points[j].y);
        let newxy={};
        newxy.x=points[j].x;
        newxy.y=points[j].y;
        if (!newxy) return true;
        me.renderShadow(newxy.x, newxy.y, points[j].value, false);
    }

    me.colorize();
};

// 随机画热点
/**
 * 随机增加热点
 * @param  {number} num 要随机生成的热力点的数量
 */
FMHeatMap.prototype.randomPoints = function (points) {

    let me = this;
    let w = me.options.width,
        h = me.options.height,
        max = me.options.points.max;

    me.clear();

    for (let i = 0; i < points; i++) {
        let x = Math.floor(Math.random() * w + 1);
        let y = Math.floor(Math.random() * h + 1);
        let value = Math.floor(Math.random() * max + 1);
        me.renderShadow(x, y, value);
    }
    me.colorize();
};

window.FMHeatMap = FMHeatMap;